#!/usr/bin/env python
'''
A CLI interface to a remote salt-api instance
'''
from __future__ import print_function

import sys
import logging

from cumin.cli import PepperCli
from cumin import PepperException

from logging import NullHandler


def main():
    logging.basicConfig(format='%(levelname)s %(asctime)s %(module)s: %(message)s')
    logger = logging.getLogger('pepper')
    logger.addHandler(NullHandler())

    try:
        cli = PepperCli()
        for exit_code, result in cli.run():
            print(result)
            if exit_code is not None:
                sys.exit(exit_code)
    # except PepperException as exc:
    #     sys.exit('Pepper error: {0}'.format(exc))
    # except Exception as e:
    #     logger.debug('Uncaught traceback:', exc_info=True)
    #     raise
    except (KeyboardInterrupt, SystemExit):
        sys.exit()


if __name__ == '__main__':
    main()



"""
A mid-level client to make executing commands easier.
"""
from collections import ChainMap
from .api import SaltApi
from .config import standard_configuration, NullCache


def _dict_filter_none(**kwarg):
    return {k: v for k, v in kwarg.items() if v is not None}


class Client:
    def __init__(self, api_url=None, *, config=None, cache=None, auto_login=True):
        """
        * api_url: URL to use, defaulting to one loaded from configuration
        * config: A configuration, defaulting to standard_configuration
        * cache: Authentication Cache to use, defaulting to NullCache
        * auto_login: Attempt to login automatically, if credentials are available
          and nothing is cached (Default: True)
        """
        self.config = config or standard_configuration()
        self.cache = cache or NullCache(self.config)
        self.api = SaltApi(api_url or self.config['url'], cache=self.cache, ssl_verify=self.config['verify'])

        if auto_login and self.config['user'] and not self.api.auth:
            self.login(self.config['user'], self.config['password'], self.config['eauth'])

    def login(self, username, password, eauth):
        return self.api.login(username, password, eauth)

    def logout(self):
        return self.api.logout()

    def events(self):
        yield from self.api.events()

    def local(self, tgt, fun, arg=None, kwarg=None, tgt_type='glob',
              timeout=None, ret=None):
        """
        Run a single execution function on one or more minions and wait for the
        results.
        """
        return self.api.run([_dict_filter_none(
            client='local',
            tgt=tgt,
            fun=fun,
            arg=arg,
            kwarg=kwarg,
            tgt_type=tgt_type,
            timeout=timeout,
            ret=ret,
        )])['return'][0]

    def local_async(self, tgt, fun, arg=None, kwarg=None, tgt_type='glob',
                    timeout=None, ret=None):
        """
        Run a single execution function on one or more minions and a generator
        producing (mid, result) pairs as they are available. (Or (None, None) if
        no new minions have responded.)

        NOTE: Every loop through the generator is an API call.
        """
        body = self.api.run([_dict_filter_none(
            client='local_async',
            tgt=tgt,
            fun=fun,
            arg=arg,
            kwarg=kwarg,
            tgt_type=tgt_type,
            timeout=timeout,
            ret=ret,
        )])
        jid = body['return'][0]['jid']
        minions = body['return'][0]['minions']

        def asynciter():
            waiting_for = set(minions)
            while waiting_for:
                # Note: runner:jobs.lookup_jid gives this to us directly, but
                # requires runner permissions
                status = self.api.jobs(jid)['info'][0]
                results = status['Result']
                finished = set(waiting_for).intersection(set(results.keys()))
                if finished:
                    for m in finished:
                        yield m, results[m]
                    waiting_for -= finished
                else:
                    yield None, None

        return minions, asynciter()

    def local_batch(self, tgt, fun, arg=None, kwarg=None, tgt_type='glob',
                    batch='50%', ret=None):
        """
        Run a single execution function on one or more minions in staged batches,
        waiting for the results.
        """
        # We don't have the option to get results as they finish, so just merge
        # everything
        batches = self.api.run([_dict_filter_none(
            client='local_batch',
            tgt=tgt,
            fun=fun,
            arg=arg,
            kwarg=kwarg,
            tgt_type=tgt_type,
            batch=batch,
            ret=ret,
        )])['return']
        return ChainMap(*batches)

    def runner(self, fun, arg=None, kwarg=None):
        """
        Run a single runner function on the master.
        """
        return self.api.run([_dict_filter_none(
            client='runner',
            fun=fun,
            arg=arg,
            kwarg=kwarg,
        )])['return'][0]

    def wheel(self, fun, arg=None, kwarg=None):
        """
        Run a single wheel function on the master.
        """
        return self.api.run([_dict_filter_none(
            client='wheel',
            fun=fun,
            arg=arg,
            kwarg=kwarg,
        )])['return'][0]



'''
A CLI interface to a remote salt-api instance

'''
import json
import logging
import argparse
import time
import sys

from .client import Client
from .config import FileCache, Config, load_config_environ, load_config_pepperrc, load_config_tui
from . import __version__

logger = logging.getLogger('pepper')


def _guess_client():
    cmd = sys.argv[0]
    if cmd.endswith('-run'):
        return 'runner'
    else:
        return 'local'


class PepperCli(object):
    def __init__(self, seconds_to_wait=3):
        self.seconds_to_wait = seconds_to_wait
        self.parser = self.get_parser()
        self.add_globalopts()
        self.add_tgtopts()
        self.add_authopts()

    def get_parser(self):
        parser = argparse.ArgumentParser(
            description=__doc__)
        parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
        return parser

    def parse(self):
        '''
        Parse all args
        '''
        self.parser.add_argument(
            '-c', dest='config', default=None,
            help='Configuration file location. Default is a file path in the '
                 '"PEPPERRC" environment variable or ~/.pepperrc.',
        )

        self.parser.add_argument(
            '-v', dest='verbose', default=0, action='count',
            help='Increment output verbosity; may be specified multiple times',
        )

        self.options = self.parser.parse_args()

    def add_globalopts(self):
        '''
        Misc global options
        '''
        optgroup = self.parser.add_argument_group(
            "Pepper ``salt`` Options", "Mimic the ``salt`` CLI")

        optgroup.add_argument('cmd', nargs='*', help='Command to run')

        optgroup.add_argument(
            '-t', '--timeout', dest='timeout', type=int, default=60,
            help='Specify wait time (in seconds) before returning control to the shell',
        )

        optgroup.add_argument(
            '--client', dest='client', default=_guess_client(),
            choices=['local', 'local_async', 'local_batch', 'runner', 'wheel'],
            help='specify the salt-api client to use (local, local_async, runner, etc)')

        optgroup.add_argument(
            '--json', dest='json_input',
            help='Enter JSON at the CLI instead of positional (text) arguments. This '
            'is useful for arguments that need complex data structures. '
            'Specifying this argument will cause positional arguments to be '
            'ignored.',
        )

        optgroup.add_argument(
            '--events', dest='events', action='store_true',
            help='Show an event stream. No commands will be executed.',
        )
        # optgroup.add_argument('--out', '--output', dest='output',
        #        help="Specify the output format for the command output")

        # optgroup.add_argument('--return', default='', metavar='RETURNER',
        #    help="Redirect the output from a command to a persistent data store")

        optgroup.add_argument(
            '--fail-if-incomplete', action='store_true', dest='fail_if_minions_dont_respond',
            default=False,
            help='Return a failure exit code if not all minions respond. This option '
            'requires the authenticated user have access to run the '
            '`jobs.list_jobs` runner function.',
        )

        return optgroup

    def add_tgtopts(self):
        '''
        Targeting
        '''
        optgroup = self.parser.add_argument_group(
            "Targeting Options",
            "Target which minions to run commands on"
        )

        optgroup.set_defaults(tgt_type='glob')

        optgroup.add_argument(
            '-E', '--pcre', dest='tgt_type', action='store_const', const='pcre',
            help="Instead of using shell globs to evaluate the target servers, "
            "use pcre regular expressions."
        )

        optgroup.add_argument(
            '-L', '--list', dest='tgt_type', action='store_const', const='list',
            help="Instead of using shell globs to evaluate the target servers, "
            "take a comma or space delimited list of servers."
        )

        optgroup.add_argument(
            '-G', '--grain', dest='tgt_type', action='store_const', const='grain',
            help='Instead of using shell globs to evaluate the target use a '
            'grain value to identify targets, the syntax for the target is the '
            'grain key followed by a globexpression: "os:Arch*".'
        )

        optgroup.add_argument(
            '-P', '--grain-pcre', dest='tgt_type', action='store_const', const='grain_pcre',
            help='Instead of using shell globs to evaluate the target use a '
            'grain value to identify targets, the syntax for the target is the '
            'grain key followed by a pcre regular expression: "os:Arch.*".'
        )

        optgroup.add_argument(
            '-N', '--nodegroup', dest='tgt_type', action='store_const', const='nodegroup',
            help="Instead of using shell globs to evaluate the target use one of "
            "the predefined nodegroups to identify a list of targets."
        )

        optgroup.add_argument(
            '-R', '--range', dest='tgt_type', action='store_const', const='range',
            help="Instead of using shell globs to evaluate the target use a range "
            "expression to identify targets. Range expressions look like %%cluster."
        )

        optgroup.add_argument(
            '-C', '--compound', dest='tgt_type', action='store_const', const='compound',
            help="The compound target option allows for multiple target types to "
            "be evaluated, allowing for greater granularity in target matching. "
            "The compound target is space delimited, targets other than globs are "
            "preceded with an identifier matching the specific targets argument "
            "type: salt 'G@os:RedHat and webser* or E@database.*'."
        )

        optgroup.add_argument(
            '-I', '--pillar', dest='tgt_type', action='store_const', const='pillar',
            help='Instead of using shell globs to evaluate the target use a pillar '
            'value to identify targets, the syntax for the target is the pillar '
            'key followed by a glob expression: "role:production*".'
        )

        optgroup.add_argument(
            '-J', '--pillar-pcre', dest='tgt_type', action='store_const', const='pillar_pcre',
            help='Instead of using shell globs to evaluate the target use a pillar '
            'value to identify targets, the syntax for the target is the pillar '
            'key followed by a pcre regular expression: "role:prod.*".'

        )

        optgroup.add_argument(
            '-S', '--ipcidr', dest='tgt_type', action='store_const', const='ipcidr',
            help="Match based on Subnet (CIDR notation) or IP address."
        )

        optgroup.add_argument('--batch', dest='batch', default=None)

        return optgroup

    def add_authopts(self):
        '''
        Authentication options
        '''
        optgroup = self.parser.add_argument_group(
            "Authentication Options",
            'Authentication credentials can optionally be supplied via the environment '
            'variables: SALTAPI_URL, SALTAPI_USER, SALTAPI_PASS, SALTAPI_EAUTH.'
        )

        optgroup.add_argument(
            '-u', '--saltapi-url', dest='saltapiurl',
            help="Specify the host url.  Defaults to https://localhost:8080"
        )

        optgroup.add_argument(
            '-a', '--auth', '--eauth', '--extended-auth', dest='eauth',
            help='Specify the external_auth backend to authenticate against and interactively '
            'prompt for credentials'
        )

        optgroup.add_argument(
            '--username', dest='username',
            help="Optional, defaults to user name. will be prompt if empty unless "
            "--non-interactive"
        )

        optgroup.add_argument(
            '--password', dest='password',
            help="Optional, but will be prompted unless --non-interactive"
        )

        optgroup.add_argument(
            '--non-interactive', action='store_false', dest='interactive', default=True,
            help="Optional, fail rather than waiting for input"
        )

        optgroup.add_argument(
            '-T', '--make-token', default=False, dest='mktoken', action='store_true',
            help="Generate and save an authentication token for re-use. The token is "
            "generated and made available for the period defined in the Salt Master."
        )

        optgroup.add_argument(
            '-x', dest='cache', default=None,
            help='Cache file location. Default is a file path in the "PEPPERCACHE" '
            'environment variable or ~/.peppercache.'
        )

        return optgroup

    CONFIG_MAP = {
        # argparse : config
        'saltapiurl': 'url',
        'eauth': 'eauth',
        'username': 'user',
        'password': 'password',
        'cache': 'cache',
    }

    def load_config_cache(self):
        config = Config()
        load_config_pepperrc(config, self.options.config)
        load_config_environ(config)

        for arg, conf in self.CONFIG_MAP.items():
            if getattr(self.options, arg, None):
                config[conf] = getattr(self.options, arg, None)

        load_config_tui(config)

        if self.options.mktoken:
            cache = FileCache(config)
        else:
            cache = None

        return config, cache

    def parse_target(self):
        opts = {}

        if self.options.events:
            return opts

        # Soooo... it turns out salt-api parses out kwargs for us, sometimes
        if self.options.client in ('local', 'local_async', 'local_batch'):
            opts.update({
                'tgt': self.options.cmd[0],
                'fun': self.options.cmd[1],
                'arg': self.options.cmd[2:],
                'tgt_type': self.options.tgt_type,
            })
        else:
            opts.update({
                'fun': self.options.cmd[0],
                'arg': self.options.cmd[1:],
            })
        if self.options.client == 'local_batch':
            opts.update({
                'batch': self.options.batch,
            })
        if self.options.client in ('local', 'local_async'):
            opts.update({
                'timeout': self.options.timeout,
            })
        return opts

    def format_response(self, data):
        return json.dumps(data, indent=4)

    def run(self):
        '''
        Parse all arguments and call salt-api
        '''
        self.parse()

        # move logger instantiation to method?
        logger.addHandler(logging.StreamHandler())
        logger.setLevel(max(logging.ERROR - (self.options.verbose * 10), 1))

        config, cache = self.load_config_cache()

        args = self.parse_target()

        self.client = Client(
            config=config,
            cache=cache,
            auto_login=True,
        )

        if self.options.json_input:
            data = json.loads(self.options.json_input)
            res = self.client.api.run(data)
            yield None, self.format_response(res)
        elif self.options.events:
            for ev in self.client.events():
                yield None, self.format_response(ev)
        elif self.options.client == 'local_async':
            minions, results = self.client.local_async(**args)
            start = time.time()
            end = start + self.options.timeout
            for mid, res in results:
                if mid is not None:
                    minions.remove(mid)
                    yield None, self.format_response({mid: res})
                    if not minions:
                        break
                if time.time() > end:
                    break
            if minions:
                ret = 1 if self.options.fail_if_minions_dont_respond else 0
                yield ret, "No response from {}".format(', '.join(minions))
        else:
            res = getattr(self.client, self.options.client)(**args)
            yield None, self.format_response(res)



"""
A low-level API exposing the salt-api HTTP calls fairly directly.
"""
import json
import logging
import urllib.parse as urlparse
import posixpath as urlpath
import requests
import tarfile
import io
from .config import NullCache
from .sse import stream_sse

logger = logging.getLogger('pepper')


class SaltTokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        super().__init__()
        self.token = token

    def __call__(self, request):
        if self.token:
            request.headers.setdefault('X-Auth-Token', self.token)
        return request


class PepperException(Exception):
    pass


class AuthenticationDenied(PepperException):
    """
    401:Authentication denied
    """


class ServerError(PepperException):
    """
    500:Server error
    """


class SaltApi(object):
    '''
    A thin wrapper for making HTTP calls to the salt-api rest_cherrpy REST
    interface

    >>> api = SaltApi('https://localhost:8000')
    >>> api.login('saltdev', 'saltdev', 'pam')
    {"return": [
            {
                "eauth": "pam",
                "expire": 1370434219.714091,
                "perms": [
                    "test.*"
                ],
                "start": 1370391019.71409,
                "token": "c02a6f4397b5496ba06b70ae5fd1f2ab75de9237",
                "user": "saltdev"
            }
        ]
    }
    >>> api.run([{'client': 'local', 'tgt': '*', 'fun': 'test.ping'}])
    {u'return': [{u'ms-0': True,
              u'ms-1': True,
              u'ms-2': True,
              u'ms-3': True,
              u'ms-4': True}]}

    '''

    def __init__(self, api_url, *, cache=None, ssl_verify=False):
        '''
        Initialize the class with the URL of the API

        :param api_url: Host or IP address of the salt-api URL;
            include the port number

        :param ssl_verify: A bool or string pointing to something that looks like a CA or trust store

        :raises ValueError: if the api_url is misformed

        '''
        split = urlparse.urlsplit(api_url)
        if split.scheme not in ['http', 'https']:
            raise ValueError("salt-api URL missing HTTP(s) protocol: {0}".format(api_url))

        if cache is None:
            self.authcache = NullCache(None)
        else:
            self.authcache = cache

        self.api_url = api_url
        self._ssl_verify = ssl_verify
        self.auth = self.authcache.get_auth() or {}
        self.session = requests.Session()

    def _construct_url(self, path):
        '''
        Construct the url to salt-api for the given path

        Args:
            path: the path to the salt-api resource

        >>> api = Pepper('https://localhost:8000/salt-api/')
        >>> api._construct_url('/login')
        'https://localhost:8000/salt-api/login'
        '''

        relative_path = path.lstrip('/')
        return urlparse.urljoin(self.api_url, relative_path)

    def _find_auth(self, data):
        eauth = data['eauth'] if data is not None and 'eauth' in data else self.auth.get('eauth')
        if eauth == 'kerberos':
            from requests_kerberos import HTTPKerberosAuth, OPTIONAL
            return HTTPKerberosAuth(mutual_authentication=OPTIONAL)
        elif self.auth and self.auth.get('token'):
            return SaltTokenAuth(self.auth['token'])
        # Don't do this because of the use of sessionless salt-api
        # else:
        #     raise MissingLogin

    def _mkrequest(self, method, path, data=None, headers={}, **kwargs):
        '''
        A thin wrapper around request and request_kerberos to send
        requests and return the response

        If the current instance contains an authentication token it will be
        attached to the request as a custom header.

        :rtype: response

        '''
        auth = self._find_auth(data)
        head = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-Requested-With': 'XMLHttpRequest',
        }
        head.update(headers)

        resp = getattr(self.session, method)(
            url=self._construct_url(path),
            headers=head,
            verify=self._ssl_verify,
            auth=auth,
            data=json.dumps(data),
            **kwargs
        )
        if resp.status_code == 401:
            raise AuthenticationDenied(resp.text)
        elif resp.status_code == 500:
            raise ServerError(resp.text)
        else:
            resp.raise_for_status()
            return resp

    def run(self, cmds):
        '''
        Execute a command through salt-api and return the response

        :param list cmds: a list of command dictionaries
        '''
        body = self._mkrequest('post', '/', cmds).json()
        return body

    def login(self, username, password, eauth):
        body = self._mkrequest('post', '/login', {
            'username': username,
            'password': password,
            'eauth': eauth,
        }).json()
        self.auth = body['return'][0]
        self.authcache.set_auth(self.auth)
        return self.auth

    def logout(self):
        self._mkrequest('post', '/logout').json()
        self.auth = {}

    def run_unsessioned(self, cmds):
        '''
        Execute a command through salt-api and return the response, bypassing
        the usual session mechanisms.

        Additional keyword arguments should be what's necessary for eauth. It's
        probably either:
        * username, password, eauth
        * token

        :param list cmds: a list of command dictionaries
        '''
        return self._mkrequest('post', '/run', cmds).json()

    def minions(self, mid):
        if mid is ...:
            path = '/minions'
        else:
            path = urlpath.join('/minions', mid)
        return self._mkrequest('get', path).json()

    def run_async(self, cmds):
        '''
        Start an execution command and immediately return the job id.

        lowstate data describing Salt commands must be sent in the request body.
        The client option will be set to local_async.

        :param list cmds: a list of command dictionaries
        '''
        return self._mkrequest('post', '/minions', cmds).json()

    def jobs(self, jid):
        if jid is ...:
            path = '/jobs'
        else:
            path = urlpath.join('/jobs', jid)
        return self._mkrequest('get', path).json()

    def keys(self, mid):
        if mid is ...:
            path = '/keys'
        else:
            path = urlpath.join('/keys', mid)
        return self._mkrequest('get', path).json()

    def key_gen(self, mid, **kwargs):
        """
        * mid: The name of the minion for which to generate a key pair.
        * keysize: The size of the key pair to generate. The size must be 2048,
          which is the default, or greater. If set to a value less than 2048,
          the key size will be rounded up to 2048.
        * force: If a public key has already been accepted for the given minion
          on the master, then the gen_accept function will return an empty
          dictionary and not create a new key. This is the default behavior. If
          force is set to True, then the minion's previously accepted key will
          be overwritten.
        * username
        * password
        * eauth
        """
        form = {
            'mid': mid,
        }
        form.update(kwargs)
        resp = self._mkrequest('post', '/keys', form)
        buf = io.BytesIO(resp.binary)
        return tarfile.open(fileobj=buf, mode='r')

    def hook(self, path, body):
        hookpath = urlpath.join('/hook', path)
        self._mkrequest('post', hookpath, body)

    def stats(self):
        return self._mkrequest('get', '/stats').json()

    def events(self):
        """
        Generator tied to the Salt event bus. Produces data roughly in the form of:

            {
                'data': {
                    '_stamp': '2017-07-31T20:32:29.691100',
                    'fun': 'runner.manage.status',
                    'fun_args': [],
                    'jid': '20170731163229231910',
                    'user': 'astro73'
                },
                'tag': 'salt/run/20170731163229231910/new'
            }

        """
        for msg in stream_sse(self._mkrequest, 'get', '/events'):
            data = json.loads(msg['data'])
            yield data



#!/usr/bin/env python
'''
A CLI front-end to a running salt-api system

'''
import json
import os
import re

from setuptools import setup
from setuptools.command import sdist

setup_kwargs = {
    'name': 'salt-cumin',
    'description': __doc__.strip(),
    'author': 'Seth House',
    'author_email': 'shouse@saltstack.com',
    'maintainer': 'Jamie Bliss',
    'maintainer_email': 'jamie.bliss@astro73.com',
    'url': 'https://github.com/astronouth7303/cumin',
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
    ],
    'packages': [
        'cumin',
    ],
    'package_data': {
        'cumin': ['version.json'],
    },
    'entry_points': {
        'console_scripts': [
            'cumin = cumin.__main__:main',
            'cumin-run = cumin.__main__:main'
        ]
    },
    'install_requires': [
        'requests',
    ],
    'extras_require': {
        'kerberos': ['requests_kerberos'],
    },
}


def versionfile(base_dir):
    return os.path.join(base_dir, 'cumin', 'version.json')


def readmefile(base_dir):
    return os.path.join(base_dir, 'README.rst')


def read_version_tag():
    git_dir = os.path.join(os.path.dirname(__file__), '.git')

    if os.path.isdir(git_dir):
        import subprocess

        try:
            p = subprocess.Popen(['git', 'describe'],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
        except Exception:
            pass
        else:
            return out.strip() or None

    else:
        return read_version_from_json_file()

    return None


def read_version_from_json_file():
    with open(versionfile(os.path.dirname(__file__))) as f:
        return json.load(f)['version']


def read_readme():
    with open(readmefile(os.path.dirname(__file__))) as f:
        return f.read()


def parse_version_tag(tag):
    '''
    Parse the output from Git describe

    Returns a tuple of the version number, number of commits (if any), and the
    Git SHA (if available).
    '''
    if isinstance(tag, bytes):
        tag = tag.decode()
    if not tag or '-g' not in tag:
        return tag, None, None

    match = re.search('(?P<version>.*)-(?P<num_commits>[0-9]+)-g(?P<sha>[0-9a-fA-F]+)', tag)

    if not match:
        return tag, None, None

    match_dict = match.groupdict()
    return (
        match_dict.get('version'),
        match_dict.get('num_commits'),
        match_dict.get('sha'))


def get_version():
    '''
    Return a tuple of the version and Git SHA
    '''
    version, num_commits, sha = parse_version_tag(read_version_tag())
    if sha:
        version = '{0}.dev{1}'.format(version, num_commits)
    return version, sha


def write_version_file(base_dir):
    ver_path = versionfile(base_dir)
    version, sha = get_version()

    with open(ver_path, 'wt') as f:
        json.dump({'version': version, 'sha': sha}, f)


class PepperSdist(sdist.sdist):
    '''
    Write the version.json file to the sdist tarball build directory
    '''

    def make_release_tree(self, base_dir, files):
        sdist.sdist.make_release_tree(self, base_dir, files)
        write_version_file(base_dir)


if __name__ == '__main__':
    version, sha = get_version()
    readme = read_readme()

    setup(cmdclass={
        'sdist': PepperSdist,
    }, version=version, long_description=readme, **setup_kwargs)  # git_sha=sha



"""
Handles configuration, credentials, caching, etc.
"""
import os
import abc
import json
import time
import getpass
import collections.abc
from configparser import ConfigParser
from .utils import umask

DEFAULT_SETTINGS = {
    'cache': os.path.expanduser('~/.peppercache'),
    'url': 'https://localhost:8000/',
    'user': None,
    'password': None,
    'eauth': 'auto',
    'verify': True
}


class AbstractCache(abc.ABC):
    def __init__(self, config):
        self.config = config

    @abc.abstractmethod
    def get_auth(self):
        """
        Loads the auth dictionary from cache, or None.

        The data is in the form of:
            {
                "eauth": "pam",
                "expire": 1370434219.714091,
                "perms": [
                    "test.*"
                ],
                "start": 1370391019.71409,
                "token": "c02a6f4397b5496ba06b70ae5fd1f2ab75de9237",
                "user": "saltdev"
            }

        This may potentially block on IO, depending on the backend used, but
        should never ask the user for information.
        """

    @abc.abstractmethod
    def set_auth(self, data):
        """
        Saves data from an auth dictionary to the cache, or None to clear it.
        Format should be the same as get_auth().
        """


class NullCache(AbstractCache):
    """
    Implements the cache interface, but does nothing.
    """

    def get_auth(self):
        return None

    def set_auth(self, data):
        pass


class FileCache(AbstractCache):
    """
    Handles caching the credentials in a file (default is ~/.peppercache)
    """

    def __init__(self, config):
        super().__init__(config)
        self.token_file = config['cache']

    def get_auth(self):
        if self.token_file and os.path.exists(self.token_file):
            with open(self.token_file, 'rt') as f:
                try:
                    auth = json.load(f)
                except json.decoder.JSONDecodeError:
                    # Assuming the file is corrupt. Eating the exception
                    return
            if auth['expire'] < time.time() + 30:  # XXX: Why +30?
                return
            return auth

    def set_auth(self, auth):
        # A bunch of extra work to set file permissions without having a window of leak
        with umask(0):
            fdsc = os.open(self.token_file, os.O_WRONLY | os.O_CREAT, 0o600)
            with os.fdopen(fdsc, 'wt') as f:
                json.dump(auth, f)


class Config(collections.abc.MutableMapping):
    """
    Configuration that just initializes itself from default values.
    """

    def __init__(self):
        self._init_from_defaults()

    def _init_from_defaults(self):
        self._data = DEFAULT_SETTINGS.copy()

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, data):
        self._data[key] = data

    def __delitem__(self, key):
        if key in DEFAULT_SETTINGS:
            self._data[key] = DEFAULT_SETTINGS[key]
        else:
            del self._data[key]

    def __iter__(self):
        yield from self._data

    def __len__(self):
        return len(self._data)


def load_config_pepperrc(config, filename=None):
    """
    Loads configurations from a pepperrc file (default is ~/.pepperrc),
    overridable by environment variables.
    """

    CONFIG_MAP = {
        'saltapi_user': 'user',
        'saltapi_pass': 'password',
        'saltapi_eauth': 'eauth',
        'saltapi_url': 'url',
        'saltapi_ssl_verify': 'verify',
    }

    if not filename:
        filename = os.path.expanduser('~/.pepperrc')

    cp = ConfigParser(interpolation=None)
    cp.read(filename)
    if 'main' in cp:
        for k, v in cp['main'].items():
            if k in CONFIG_MAP:
                config[CONFIG_MAP[k]] = v


def load_config_environ(config, environ=None):
    """
    Loads configurations from process environmnet (default os.environ)
    """

    CONFIG_MAP = {
        'SALTAPI_USER': 'user',
        'SALTAPI_PASS': 'password',
        'SALTAPI_EAUTH': 'eauth',
        'SALTAPI_URL': 'url',
        'SALTAPI_SSL_VERIFY': 'verify',
        'PEPPERCACHE': 'cache',
    }

    if environ is None:  # Don't overwrite {}
        environ = os.environ

    for k, v in environ.items():
        if k in CONFIG_MAP:
            config[CONFIG_MAP[k]] = v


def load_config_tui(config):
    """
    Prompts the user for information.
    """

    PROMPTERS = {
        'user': lambda: input('Username: '),
        'password': (
            lambda:
            None if config['eauth'] == 'kerberos' else getpass.getpass(prompt='Password: ')
        ),
    }

    for field, prompter in PROMPTERS.items():
        if not config[field]:
            config[field] = prompter()


def standard_configuration(*, pepperrc=None, environ=None):
    """
    Builds a standard configuration, suitable for most API clients.

    Standard uses this order:
    1. Process environment
    2. .pepperrc file
    3. Hard-coded defaults
    """
    config = Config()
    load_config_pepperrc(config, pepperrc)
    load_config_environ(config, environ)
    return config



"""
A requests-based Server-sent Events client implementation
"""

import requests
import time


def stream_raw_sse(mkrequest, *pargs, _last_event_id=None, headers=None, **kwargs):
    """
    Streams Server-Sent Events, each event produced as a sequence of
    (field, value) pairs.

    Does not handle reconnection, etc.
    """
    if headers is None:
        headers = {}
    headers['Accept'] = 'text/event-stream'
    headers['Cache-Control'] = 'no-cache'
    # Per https://html.spec.whatwg.org/multipage/server-sent-events.html#sse-processing-model
    if _last_event_id is not None:
        headers['Last-Event-ID'] = _last_event_id

    with mkrequest(*pargs, headers=headers, stream=True, **kwargs) as resp:
        fields = []
        for line in resp.iter_lines(decode_unicode=True):
            # https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation
            if not line:
                yield fields
                fields = []
            elif line.startswith(':'):
                pass
            elif ':' in line:
                field, value = line.split(':', 1)
                if value.startswith(' '):
                    value = value[1:]
                fields += [(field, value)]
            else:  # Non-blank, without a colon
                fields += [(line, '')]


def stream_sse(mkrequest, *pargs, **kwargs):
    """
    Streams server-sent events, producing a dictionary of the fields.

    Handles reconnecting, Last-Event-ID, and retry waits.

    Deviates by spec by passing through unknown fields instead of ignoring them.
    If an unknown field is given more than once, the last given wins (like
    event and id).
    """
    retry = 0
    last_id = None
    while True:
        try:
            for rawmsg in stream_raw_sse(mkrequest, *pargs, _last_event_id=last_id, **kwargs):
                msg = {'event': 'message', 'data': ''}
                # https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation
                for k, v in rawmsg:
                    if k == 'retry':
                        try:
                            retry = int(v)
                        except ValueError:
                            pass
                    elif k == 'data':
                        if msg['data']:
                            msg['data'] += '\n' + v
                        else:
                            msg['data'] = v
                    else:
                        if k == 'id':
                            last_id = v
                        # Spec says we should ignore unknown fields. We're passing them on.
                        msg[k] = v
                if not msg['data']:
                    pass
                yield msg
            else:
                raise StopIteration  # Really just exists to get caught in the next line
        except (StopIteration, requests.RequestException, EOFError):
            # End of stream, try to reconnect
            # NOTE: GeneratorExit is thrown if the consumer kills us (or we get GC'd)
            # TODO: Log something?

            # Wait, fall through, and start at the top
            time.sleep(retry / 1000)
